//
// C++ Interface: eel_util
//
// Description: 
//
//
// Author: Ryan Cunningham <rcunning@cs.ucla.edu>, (C) 2004
//
// Copyright: Ryan Cunningham
//
//

#ifndef EEL_UTIL_H
#define EEL_UTIL_H

// include macro helpers here and only here
#include "eel_macro_helpers.h"
#include "eel_macro_types.h"

#include <queue.h>
#include <tree.h>

//
// eel core structs
//
// event types
typedef enum { EEL_READ_EVENT = 0, EEL_WRITE_EVENT, EEL_ERROR_EVENT, EEL_TIMER_EVENT } eel_event_type;
// state changes from left to right, always (but can skip waiting to expired)
typedef enum { EEL_CREATED = 0, EEL_WAITING, EEL_FIRE, EEL_EXPIRED } eel_state;
// the event structure
struct eel_event {
  // data items
  eel_event_type   type;          // event type
  eel_state        state;         // event state
  eel_callback     callback;      // the callback function for this event
  void             *callback_arg; // the callback arg for this event
  struct eel_group *group;        // the 'connection' group identifier
  union {                         // fd or time
    int            fd;
    struct timeval timer;
  };
  // Q structs, queue for all, fired_queue for events that fired
  TAILQ_ENTRY(eel_event)   queue_entry;
  TAILQ_ENTRY(eel_event)   fire_queue_entry;
  // timer min Red/Black tree node
  RB_ENTRY(eel_event)      timer_rb_entry;
  // group entry
  TAILQ_ENTRY(eel_event)   group_queue_entry;
  // link to ORs
  RING_ENTRY(eel_event)    OR_ring;
};

//
// create the queue types and any prototypes necessary
//
TAILQ_HEAD(event_tail_queue, eel_event);
RB_HEAD(event_rb_tree, eel_event);
RB_PROTOTYPE(event_rb_tree, eel_event, timer_rb_entry, compare_time_values);

// the group structure and queue type
struct eel_group {
  TAILQ_ENTRY(eel_group)   group_queue_entry;
  struct event_tail_queue  event_queue;
};
TAILQ_HEAD(group_tail_queue, eel_group);

//
// eel operators structure - implement to provide new
// wait mechanism
//
struct eel_lib_struct;
struct eel_operators
{
  char  *name;
  int   (*init)  ();
  void  (*end)   ();
  int   (*add)   (struct eel_lib_struct *eel, struct eel_event *e);
  int   (*remove)(struct eel_lib_struct *eel, struct eel_event *e);
  int   (*wait)  (struct eel_lib_struct *eel, int timeout_milliseconds);
};

//
// the eel library structure
//
struct eel_lib_struct
{
  // data structures
  struct event_tail_queue   queue;
  struct event_tail_queue   fire_queue;
  struct event_rb_tree      timer_rb;
  struct group_tail_queue   group_queue;
  // internal data fields
  int volatile              exit_loop;
  int                       exit_on_signal;
  // general operators
  int (*add_to_fire)(struct eel_event *event);
  // wait operators structure
  struct eel_operators *ops;
};

//
// helper function to get the perfered eel_operator implmentation
//
extern struct eel_operators *get_preferred_eel_operators();

#endif // EEL_UTIL_H
